// Everyone's favorite backend. Copyright 2011, Tyler Washburn, MIT License.

"use strict"

var basics, configurations, data, events, fs, http, markdown, path, qs, server, trailer, url, rid, zlib;

Function.prototype.is = function ( item ) { return item != null && ( item.constructor === this.prototype.constructor ) }

events = require( "events" ),
fs = require( "fs" ),
http = require( "http" ),
path = require( "path" ),
qs = require( "querystring" ),
url = require( "url" ),
markdown = require( "markdown" ).markdown,
zlib = require( "zlib" );

// This isn't one of our event emitters,
// because then it would emit to itself in an infinite loop.
global.trailer = trailer = module.exports = new events.EventEmitter()
trailer.version = "0.7"

// We put the http module here so that it can be extended.
trailer.http = http

trailer.EventEmitter = function () {
    var mark, prop;
    
    mark = new events.EventEmitter()
    
    for ( prop in mark ) ( function ( method, name ) {
        if ( Function.is( events.EventEmitter.prototype[ name ] ) ) {
            mark[ name ] = function () {
                trailer[ name ].apply( trailer, arguments )
                return method.apply( mark, arguments )
            }
        }
    } )( mark[ prop ], prop )
    
    return mark
}

trailer.hitch = function ( root, conf ) {
    var prop, indexes;

    !conf&&( conf = {} )
    
    if ( !String.is( root ) ) {
        conf = root
        root = conf.root
    } else {
        conf.root = root
    }
    
    !conf.headers&&( conf.headers = {} )
    
    for ( prop in basics.headers ) if ( basics.headers.hasOwnProperty( prop ) && !conf.headers.hasOwnProperty( prop ) ) {
        conf.headers[ prop ] = basics.headers[ prop ]
    }
    
    if ( conf.mimes !== false ) {
        !conf.mimes&&( conf.mimes = {} )
            
        for ( prop in basics.mimes ) if ( basics.mimes.hasOwnProperty( prop ) && !conf.mimes.hasOwnProperty( prop ) ) {
            conf.mimes[ prop ] = basics.mimes[ prop ]
        }
    }
    
    if ( Array.is( conf.indexes ) ) {
        indexes = conf.indexes
        conf.indexes = {}
        indexes.forEach( function ( index ) {
            conf.indexes[ index ] = 0
        } )
    }
    
    if ( conf.host ) {
        configurations[ conf.host ] = conf
    } else if ( conf.hosts ) conf.hosts.forEach( function ( host ) {
        configurations[ host ] = conf
    } )
    
    !configurations.fallback&&( configurations.fallback = conf )
    !server._handle&&( server.listen( Number( conf.port ) || 80 ) )
    
    conf._emitter = new trailer.EventEmitter()
    
    return conf._emitter
}

server = trailer.server = http.createServer(),
basics = trailer.basics = {
    "headers": {
        "Server": "Trailer/" + trailer.version + " Node/" + process.versions.node
    },
    
    "mimes": {
        "*": "text/plain"
    }
},
configurations = {},
data = trailer.data = {
    "types": {
        "application/x-www-form-urlencoded": function ( data ) {
            return qs.parse( data )
        },
        
        "application/json": function ( data ) {
            return JSON.parse( data )
        }
    },
    
    "process": function ( type, given ) {
        return Function.is( data.types[ type ] ) ? data.types[ type ]( given ) : {}
    }
};

http.ServerResponse.prototype.error = function ( code ) {
    var i, o;
    
    i = this.request,
    o = this,
    o.status = http.STATUS_CODES[ code ] ? code : 500;
    
    if ( i.conf.errors && i.conf.errors[ code ] ) {
        if ( Function.is( i.conf.errors[ code ] ) ) {
            i.conf.errors[ code ]( i, o )
        } else {
            path.exists( i.conf.errors[ code ], function ( exists ) {
                if ( exists ) {
                    fs.readFile( i.conf.errors[ code ], function ( error, content ) {
                        o.finish( content )
                    })
                } else {
                    path.exists( path.join( i.conf.root, i.conf.errors[ code ] ), function ( exists ) {
                        if ( exists ) {
                            fs.readFile( path.join( i.conf.root, i.conf.errors[ code ] ), function ( error, content ) {
                                o.finish( content )
                            })
                        } else {
                            o.finish( i.conf.errors[ code ] )
                        }
                    })
                }
            })
        }
    } else {
        o.finish( code + ": " + http.STATUS_CODES[ code ] )
    }
    
    return this
}

http.ServerResponse.prototype.extend = function () {
    var i, prop;
    
    for ( i = 0; i < arguments.length; i++ ) {
        if ( Number.is( arguments[ i ] ) ) {
            this.status = arguments[ i ]
        } else if ( String.is( arguments[ i ] ) || Buffer.is( arguments[ i ] ) || arguments[ i ] === null ) {
            this.body = arguments[ i ]
        } else {
            for ( prop in arguments[ i ] ) {
                if ( arguments[ i ].hasOwnProperty( prop ) ) {
                    this.headers[ prop ] = arguments[ i ][ prop ]
                }
            }
        }
    }
    
    return this
}

http.ServerResponse.prototype.proceed = function () {
    this.extend.apply( this, arguments )
    
    this.writeHead( this.status, this.headers )
    this.body != null && this.stream.write( this.body )
    
    return this
}

http.ServerResponse.prototype.finish = function () {
    // arguments -> .proceed -> .extend
    this.proceed.apply( this, arguments )
    this.stream.end()
    
    return this
}

rid = 0
    
server.on( 'request', function ( i, o ) {
    i.received = new Date(),
    i.rid = rid++,
    i.host = ( /(www[0-9]*\.)*([a-z0-9\-\.]+|\[[abcdef0-9\:]+\])(\:[0-9]+)?/i.exec( i.headers.host ) || [] )[2],
    i.conf = configurations[ i.host ] || configurations.fallback,
    i.handler = i.conf[ i.method ] || trailer[ i.method ];
    
    i.response = o,
    o.request = i;
    
    var prop;
    
    if ( i.handler ) {
        i.url = url.parse( i.url ),
        i.path = path.join( i.conf.root, i.url.pathname ),
        i.cookies = qs.parse( i.headers.cookie, '; ' ),
        i.cacheDate = i.headers[ "if-modified-since" ] ? new Date( i.headers[ "if-modified-since" ] ) : null,
        i.depth = 0,
        i.up = i.url.pathname,
        i.body = "";
    
        i.on( 'data', function ( data ) {
            i.body += data
        } )
        
        o.emitter = i.conf._emitter,
        o.status = 500,
        o.headers = { "Date": i.received.toUTCString() },
        o.body = "";
        
        i.on( 'end', function () {
            i.headers["content-type"]&&( i.data = data.process( i.headers["content-type"], i.body ) )
            
            o.emitter.emit( 'received', i, o )
            i.handler( i, o )
        })
    
        for ( prop in i.conf.headers ) {
            i.conf.headers.hasOwnProperty(prop)&&( o.headers[ prop ] = i.conf.headers[ prop ] )
        }
    } else {
        o.error( 501 )
    }
})

trailer.GET = trailer.POST = trailer.HEAD = function ( i, o ) {

    if ( i.conf.gzip && /gzip/.test( i.headers[ "accept-encoding" ] ) ) {
        o.extend( { "Content-Encoding": "gzip" } )
        o.stream = zlib.createGzip()
        o.stream.pipe( o )
    } else {
        o.stream = o
    }
    
    if ( i.conf.api && Function.is( i.conf.api[ i.up ] ) ) {
        i.conf.api[ i.up ]( i, o, i.extra )
    } else if ( !i.depth && i.conf.redirect && i.conf.redirect[ i.url.pathname ] ) {
        o.extend( { "Location": i.conf.redirect[ i.url.pathname ] } )
         .error( 307 )
    } else {
        path.exists( i.path, function ( exists ) {
            if ( exists ) {
                o.status = 200
                fs.stat( i.path, function ( error, stats ) {
                    if ( i.cacheDate && ( i.cacheDate.getTime() === stats.mtime.getTime() ) ) {
                        o.status = 304
                    }
                    
                    // The client only wants headers, or has a cached version.
                    if ( i.method === "HEAD" || o.status === 304 ) {
                        o.body = null
                    }
                    
                    if ( stats.isFile() ) {
                        i.conf.mimes&&!o.headers["Content-Type"]&&( o.headers[ "Content-Type" ] = i.conf.mimes[ path.extname( i.path ) ] || i.conf.mimes[ "*" ] )
                        !o.headers["Content-Length"]&&( o.headers[ "Content-Length" ] = stats.size )
                        !o.headers["Last-Modified"]&&( o.headers[ "Last-Modified" ] = stats.mtime.toUTCString() )
                        
                        if ( o.body != null ) {
                            fs.readFile( i.path, function ( error, content ) {
                                if ( error ) {
                                    if ( error.message.match( /EACCESS/ ) ) {
                                        o.error( 403 )
                                    } else {
                                        o.error( 500 )
                                    }
                                } else {
                                    o.body = content;
                                    o.emitter.emit( "finalized", i, o )
                                }
                            })
                        } else {
                            o.emitter.emit( "finalized", i, o )
                        }
                    } else if ( stats.isDirectory() ) {
                        var indexes, file, check;
                        
                        indexes = Object.keys( i.conf.indexes );
                        
                      ( check = function check( index ) {
                            file = indexes[ index ]
                            if ( index < indexes.length ) {
                                if ( i.depth <= i.conf.indexes[ file ] ) {
                                    path.exists( path.resolve( i.path, file ), function ( exists ) {
                                        if ( exists ) {
                                            i.path = path.resolve( i.path, file )
                                            i.handler( i, o )
                                        } else {
                                            check( index + 1 )
                                        }
                                    } )
                                } else {
                                    check( index + 1 )
                                }
                            } else {
                                if ( !i.depth || i.depth === 1 && i.extra === "directory.html" ) {
                                    fs.readdir( i.path, function ( error, files ) {
                                        if ( error ) {
                                            if ( error.message.match( /EACCESS/ ) ) {
                                                o.error( 403 )
                                            } else {
                                                o.error( 500 )
                                            }
                                        } else {
                                            o.headers[ "Content-Type" ] = "text/html"
                                            files.forEach( function ( name ) {
                                                name.charAt(0)!=="."&&( o.body += "- [" + name + "](" + path.join( i.up, name ) + ")\n" )
                                            })
                                            
                                            o.body = markdown.toHTML( o.body )
                                        
                                            o.emitter.emit( "finalized", i, o )
                                        }
                                    })
                                } else if ( i.depth === 1 && i.extra === "directory.json" ) {
                                    fs.readdir( i.path, function ( error, files ) {
                                        if ( error ) {
                                            if ( error.message.match( /EACCESS/ ) ) {
                                                o.error( 403 )
                                            } else {
                                                o.error( 500 )
                                            }
                                        } else {
                                            o.extend( { "Content-Type": "application/json" },
                                                JSON.stringify( files.filter( function ( name ) {
                                                    return name.charAt( 0 ) !== "."
                                                } ) )
                                            )
                                        
                                            o.emitter.emit( "finalized", i, o )
                                        }
                                    })
                                } else {
                                    o.error( 404 )
                                }
                            }
                        } )( 0 )
                    }
                })
            } else {
                if ( i.up.length > 1 ) {
                    i.path = path.join( i.path, ".." )
                    i.up = path.join( i.up, ".." )
                    i.extra = path.relative( i.up, i.url.pathname )
                    
                    i.depth++
                    
                    i.handler( i, o )
                } else {
                    o.error( 404 )
                }
            }
        })
    }
}




